const fs = require('fs-extra')
const path = require('path')
const ora = require('ora')
const kebabCase = require('lodash.kebabcase')
const { print, run } = require('@toptal/davinci-cli-shared')

const skeleton = require('../utils/skeleton')
const { findLatestVersionOfPackage } = require('../utils/packages')

const checkPrerequisites = async (destDir, appName) => {
  const spinner = ora('Checking prerequisites...').start()

  // look for env resolvers
  try {
    await run('direnv', ['version'])
  } catch (err) {
    spinner.warn()
    print.yellow(
      "warning: You don't have `direnv` for resolving env variables! Expect other approach is used."
    )
  }

  // check if app already exists
  const doesAppExist = await fs.exists(destDir)

  if (doesAppExist) {
    spinner.fail()
    throw new Error(`Application with name ${appName} already exists!`)
  }

  spinner.succeed('Prerequisites checked!')
}

const copyApplicationSkeleton = async (destDir, currentDavinciRunningDir) => {
  const skeletonPath = require.resolve('@toptal/davinci-skeleton', {
    paths: [currentDavinciRunningDir],
  })
  const appDir = path.join(skeletonPath, '../..')

  const spinner = ora('Copying skeleton...').start()

  try {
    await skeleton.copyAllFiles(appDir, destDir)
    await skeleton.copySPAWorkflows(appDir, destDir)
    await skeleton.copyNpmrc(destDir)
    await skeleton.copyTemplatePackageJsonFile(destDir)
    await skeleton.copyTemplateGitIgnoreFile(destDir)
    await skeleton.copyTemplateEslintrc(destDir)

    // remove bin directory
    await fs.remove(path.join(destDir, 'bin'))

    spinner.succeed('Skeleton copied!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const addVersionsToPackageJson = async (packageJson, destDir) => {
  // list of dependencies to be added to a new project
  const dependencies = [
    '@toptal/picasso-provider',
    '@toptal/picasso',
    '@topkit/apollo-client',
    'react',
    'react-dom',
    'styled-components',
    'typescript',
    '@apollo/client',
    'graphql',
  ]

  const devDependencies = [
    '@toptal/davinci-qa',
    '@toptal/davinci-engine',
    '@toptal/davinci-syntax',
    '@toptal/davinci-code',
    '@toptal/davinci-ci',
    '@types/react',
    '@types/react-dom',
    '@toptal/davinci-storybook',
    '@toptal/davinci-storybook-theme',
    '@toptal/davinci-storybook-addons',
    '@toptal/davinci-storybook-decorators',
  ]

  const toNameVersionPair = async dependency => {
    const version = await findLatestVersionOfPackage(dependency, destDir)

    return [dependency, `^${version}`]
  }

  const dependenciesVersionsPromise = dependencies.map(toNameVersionPair)
  const devDependenciesVersionPromise = devDependencies.map(toNameVersionPair)

  packageJson.dependencies = {
    ...packageJson.dependencies,
    ...Object.fromEntries(await Promise.all(dependenciesVersionsPromise)),
  }

  packageJson.devDependencies = {
    ...packageJson.devDependencies,
    ...Object.fromEntries(await Promise.all(devDependenciesVersionPromise)),
  }
}

const modifyPackageJson = async (destDir, appName) => {
  const spinner = ora('Modifiying package.json...').start()

  try {
    const config = skeleton.readPackageJson(destDir)

    config.name = `@toptal/${appName}`
    config.homepage = `https://github.com/toptal/${appName}/tree/master#readme`

    delete config.gitHead

    delete config.scripts['prepack']
    delete config.scripts['postpack']
    delete config.scripts['build:package']
    delete config.scripts['prepublishOnly']

    // add husky scripts
    config.scripts.prepare = "if [ -d '.git' ]; then husky install; fi"

    delete config.publishConfig
    config.license = 'SEE LICENSE IN LICENSE.MD'

    config.version = '0.0.1'

    await addVersionsToPackageJson(config, destDir)

    skeleton.writePackageJson(destDir, config)

    spinner.succeed('package.json modified!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const installDependencies = async destDir => {
  const spinner = ora('Installing dependencies... ').start()

  try {
    await run('yarn', ['install', '--prefer-offline'], { cwd: destDir })

    spinner.succeed('Dependencies installed!')
  } catch (err) {
    spinner.fail()
    throw new Error(err)
  }
}

const createCommand = async (appName, { skipInstall }) => {
  print.green('Creating', appName, '...')

  const currentRunningDir = process.cwd()
  const currentDavinciRunningDir = __dirname
  const destDir = path.join(currentRunningDir, appName)

  try {
    // check if all apps are installed, directory doesn't exists
    await checkPrerequisites(destDir, appName)

    // copy skeleton to destination directory
    await copyApplicationSkeleton(destDir, currentDavinciRunningDir)

    // change package.json
    await modifyPackageJson(destDir, appName)

    // run yarn install
    if (!skipInstall) {
      await installDependencies(destDir)
    }

    print.green('Application has been created.')
  } catch (err) {
    print.red(err)
    process.exit(1)
  }
}

const newCommandCreator = {
  command: 'new <appName>',
  description: 'Create new application',
  options: [
    {
      name: '--skip-install',
      label: 'Skip node modules installation',
    },
  ],
  action: async ({ appName, options: { 'skip-install': skipInstall } }) => {
    const formattedAppName = kebabCase(appName)

    return createCommand(formattedAppName, { skipInstall })
  },
}

module.exports = newCommandCreator
